#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include "molTypes.h"
#include "gridTypes.h"
#include "grid.h"
#include "pdbLoader.h"
#include "sceneBuilder.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

// Timing variables
clock_t m_tTime;
int m_iNbFrames;
float m_fFPS;
clock_t ttime;

char spinFlag = 0;

float sphi=0.0, stheta=0.0;
float somega=0.0;
float sdepth = 950;
float sx = 0;
float sy = 0;
float zNear = 3.0;
float zFar = 2000.0;
float aspect = 1.0;

int lastx, lasty, currx, curry;
char isRotatingZ;
char isRotatingXY;
char isZooming;
char isPanning;

short WIDTH = 512;
short HEIGHT = 512;
short N = 128;
float PR = 1.4;
Molecule mol = NULL;
Molecule catrace = NULL;
Grid grid = NULL;

GLfloat white[] = {1.0, 1.0, 1.0, 1.0};
GLfloat gray[] = {0.6, 0.6, 0.6, 1.0};
GLfloat gray2[] = {0.25, 0.25, 0.25, 1.0};
GLfloat light_position[] = {0.5, 0.0, 2.0, 0.0};

void window_pos( GLfloat x, GLfloat y, GLfloat z, GLfloat w )
{
   GLfloat fx, fy;

   /* Push current matrix mode and viewport attributes */
   glPushAttrib( GL_TRANSFORM_BIT | GL_VIEWPORT_BIT );

   /* Setup projection parameters */
   glMatrixMode( GL_PROJECTION );
   glPushMatrix();
   glLoadIdentity();
   glMatrixMode( GL_MODELVIEW );
   glPushMatrix();
   glLoadIdentity();

   glDepthRange( z, z );
   glViewport( (int) x - 1, (int) y - 1, 2, 2 );

   /* set the raster (window) position */
   fx = x - (int) x;
   fy = y - (int) y;
   glRasterPos4f( fx, fy, 0.0, w );

   /* restore matrices, viewport and matrix mode */
   glPopMatrix();
   glMatrixMode( GL_PROJECTION );
   glPopMatrix();

   glPopAttrib();
}


void renderBitmapString( float x, float y, float z, void *font, char *string) 
{  
  char *c;
  window_pos(x,y,z,1.0);
  for (c=string; *c != '\0'; c++) 
  {
    glutBitmapCharacter(font, *c);
  }
}


void display(void)
{
	char s[20];
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable(GL_LIGHTING);

	if (spinFlag) stheta+=2.0;
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(-sx,-sy,-sdepth);
	glRotatef(stheta,1.0,0.0,0.0);
	glRotatef(-sphi,0.0,1.0,0.0);
	glRotatef(somega,0.0,0.0,1.0);
	glCallList(1);

	// Calculating frames per second
	if( m_iNbFrames++ > 25 )
	{
		m_fFPS = (float)m_iNbFrames / ( clock() - m_tTime ) * CLOCKS_PER_SEC;
		m_iNbFrames = 0;
		m_tTime = clock();
	}

	// Showing FPS
	glDisable( GL_LIGHTING );
	glColor3f( 0.0f, 0.0f, 0.0f );
	sprintf(s,"FPS: %f",m_fFPS);
	renderBitmapString(-0.5f, 0.35f, 0.0f,GLUT_BITMAP_HELVETICA_18,s);

	glFinish();
	glutSwapBuffers();
}

void idle(void)
{
//	glNewList(1,GL_COMPILE);
//	buildScene(grid);
//	glEndList();
	glutPostRedisplay();
}

void reshape(int w, int h)
{
	glViewport(0, 0, w, h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	aspect = (float)w/(float)h;
	gluPerspective(45.0, aspect, zNear, zFar);
}

void keyboard(unsigned char key, int xPos, int yPos)
{
	switch (key)
	{
		case 27:
			exit(0);
	}
}

void vpRotateXY(float x, float y)
{
	sphi+=x;
	stheta+=y;
}

void vpRotateZ(float z)
{
	somega+=z;
}

void vpZoom(float x)
{
	sdepth+=x;
}

void vpPan(float x, float y)
{
	sx+=x;
	sy+=y;
}

void mouse(int button, int state, int x, int y)
{
	int mod;

	lastx = x;
	lasty = y;
	currx = lastx;
	curry = lasty;

	mod = glutGetModifiers();

	if ((button == GLUT_LEFT_BUTTON) && (state == GLUT_DOWN) && (mod == GLUT_ACTIVE_SHIFT))
	{
		//Zooming
      		isRotatingXY=0;
      		isRotatingZ=0;
		isZooming=1;
		isPanning=0;
	}
	else if ((button == GLUT_RIGHT_BUTTON) && (state == GLUT_DOWN) && (mod == GLUT_ACTIVE_CTRL))
	{
		//Panning
		isRotatingXY = 0;
		isRotatingZ = 0;
		isZooming = 0;
		isPanning = 1;
	}
	else if ((button == GLUT_RIGHT_BUTTON) && (state == GLUT_DOWN) && (mod == GLUT_ACTIVE_SHIFT))
	{
  		//Rotating around Z axis
		isRotatingXY = 0;
		isRotatingZ = 1;
		isZooming = 0;
		isPanning = 0;
    	}
	else if ((button == GLUT_LEFT_BUTTON) && (state == GLUT_DOWN))
	{
		//Rotating
		isRotatingXY = 1;
		isRotatingZ = 0;
		isZooming = 0;
		isPanning = 0;
	}
	else 
	{
		isRotatingXY = 0;
		isRotatingZ = 0;
		isZooming = 0;
		isPanning = 0;
	}

	glutPostRedisplay();

}

void motion(int x, int y)
{
	float scaling1 = 0.25;
	float scaling2 = 2.0;

	currx = x;
	curry = y;

	if (isRotatingXY)
	{
		//Rotating XY
		vpRotateXY(-scaling1*(currx - lastx),scaling1*(curry - lasty));
	}
	else if (isRotatingZ)
	{
		//Rotating Z
		vpRotateZ(-scaling1*(currx - lastx));
	}
	else if (isZooming)
	{
		//Zooming
		vpZoom(scaling2*(curry - lasty));
	}
	else if (isPanning)
	{
		//Panning
		vpPan(-scaling2*(currx - lastx),scaling2*(curry - lasty));
	}

	lastx = currx;
	lasty = curry;

	glutPostRedisplay();
}

void menu(int item)
{
	switch(item)
	{
		case 0:
			exit(0);

		case 1:
			spinFlag = !spinFlag;
			break;
	}
}

void init(void)
{
	clock_t t;
	glClearColor(1.0, 1.0, 1.0, 0.0);

	glLightfv(GL_LIGHT0, GL_SPECULAR, white);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, gray);
	glLightfv(GL_LIGHT0, GL_AMBIENT, gray2);
	glLightfv(GL_LIGHT0, GL_POSITION, light_position);

	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_DEPTH_TEST);
	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);

	t = clock();
	glNewList(1,GL_COMPILE);
	buildScene(grid);
	buildCATrace(catrace);
	printf("Scene building time %f seconds\n",(clock()-t)/(float)CLOCKS_PER_SEC);
	t = clock();
	glEndList();
	printf("OpenGL display list compilation time %f seconds\n",(clock()-t)/(float)CLOCKS_PER_SEC);
	printf("Total Molecular Surface Generation Time : %f seconds\n",(clock()-ttime)/(float)CLOCKS_PER_SEC);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(64.0, aspect, zNear, zFar);

	m_iNbFrames = 0;
	m_tTime = clock();
	m_fFPS = 0;

}

float centerMolecule(Molecule mol, Molecule caTrace)
{
	// scale and translate the molecule so that the atom coordinates are between -250 and 250
	float minx,miny,minz,maxx,maxy,maxz;
	float s,sx,sy,sz;
	int i;

	minx = mol->xpoints[0];
	maxx = mol->xpoints[0];
	miny = mol->ypoints[0];
	maxy = mol->ypoints[0];
	minz = mol->zpoints[0];
	maxz = mol->zpoints[0];

	for (i=1;i<mol->npoints;i++){
		if (mol->xpoints[i]<minx)	minx=mol->xpoints[i];
		if (mol->xpoints[i]>maxx)	maxx=mol->xpoints[i];
		if (mol->ypoints[i]<miny)	miny=mol->ypoints[i];
		if (mol->ypoints[i]>maxy)	maxy=mol->ypoints[i];
		if (mol->zpoints[i]<minz)	minz=mol->zpoints[i];
		if (mol->zpoints[i]>maxz)	maxz=mol->zpoints[i];
	}
	sx = 400.0f/(maxx-minx);
	sy = 400.0f/(maxy-miny);
	sz = 400.0f/(maxz-minz);
	s = 0.0f;
	if (sx<sy && sx<sz)	s=sx;
	else if (sy<sx && sy<sz)s=sy;
	else s=sz;
	//printf("%f %f %f\n",sx,sy,sz);
	//printf("s = %f\n",s);
	printf("SPDBV quality = %f\n",(PR*s)/(512/N));

	for (i=0;i<mol->npoints;i++){
		mol->xpoints[i]=(s*(mol->xpoints[i]-((minx+maxx)/2)));
		mol->ypoints[i]=(s*(mol->ypoints[i]-((miny+maxy)/2)));
		mol->zpoints[i]=(s*(mol->zpoints[i]-((minz+maxz)/2)));
		mol->rpoints[i]=mol->rpoints[i]*s;
	}
	for (i=0;i<catrace->npoints;i++){
		catrace->xpoints[i]=(s*(caTrace->xpoints[i]-((minx+maxx)/2)));
		catrace->ypoints[i]=(s*(caTrace->ypoints[i]-((miny+maxy)/2)));
		catrace->zpoints[i]=(s*(caTrace->zpoints[i]-((minz+maxz)/2)));
		catrace->rpoints[i]=caTrace->rpoints[i]*s;
	}
	return s;
}

int main(int argc, char** argv)
{
	float s;
	clock_t t;
	int c;
	char inner = 0;
	float k = 0;

	if (argc!=5)
	{
		printf("Usage: levelSet <pdb File> <gridSize> <probeSize> <inner caves(1) or outer surface(0)\n");
		return 1;
	}
	t = clock();
	mol = loadMolecule(argv[1]);
	catrace = loadCATrace(argv[1]);
	printf("Number of CA atoms = %d\n",catrace->npoints);
	printf("Molecule loading time %f seconds\n",(clock()-t)/(float)CLOCKS_PER_SEC);
	N = (short)atoi(argv[2]);
	PR = (float)atof(argv[3]);
	inner = argv[4][0]-'0';
	printf("Number of atoms: %d\n",mol->npoints);
	t = clock();
	ttime = clock();
	s = centerMolecule(mol, catrace);
	printf("Molecule centering time %f seconds\n",(clock()-t)/(float)CLOCKS_PER_SEC);
	PR = PR*s;

	t = clock();
	grid = createGrid(N);
	printf("Grid creation time %f seconds\n",(clock()-t)/(float)CLOCKS_PER_SEC);
	t = clock();

	signDistanceGridMol(grid,mol,PR);
	printf("Grid initialization time %f seconds\n",(clock()-t)/(float)CLOCKS_PER_SEC);
	
	t = clock();

	shrink(grid,PR);
/*	c = findProbesMol(grid,PR);
	printf("There are %d probes.\n",c);
	printf("ProbesMol time %f seconds.\n",(clock()-t)/(float)CLOCKS_PER_SEC);
*/
	printf("Shrinking time %f seconds.\n",(clock()-t)/(float)CLOCKS_PER_SEC);
	
	
	t = clock();
	k = s/(512/N);
	k = k * k * k;
	printf("Total cavity/molecule volume = %f\n",fastMarching(grid,inner)/k);
	printf("Fast marching time %f seconds\n",(clock()-t)/(float)CLOCKS_PER_SEC);
	
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);

	glutInitWindowSize(WIDTH,HEIGHT);
	glutInitWindowPosition(100,100);
	glutCreateWindow("LSMS: Molecular Surfaces using Level Set method");
	init();
	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyboard);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutIdleFunc(idle);

	glutCreateMenu(menu);
	glutAddMenuEntry("Toggle spin", 1);
	glutAddMenuEntry("", -1);
	glutAddMenuEntry("Exit", 0);
	glutAttachMenu(GLUT_MIDDLE_BUTTON);

	glutMainLoop();
	return 0;
}
